#!/usr/bin/env python
# -*- coding: utf-8 -*-

import threading
import mraa


class ShiftRegister(object):
    """!
    Schieberegister für D_IN und D_OUT. Definitionen der Schieberegister Pins.
    """
    # Outputs
    # default bit positions
    V24_ON = 0
    LED_GRN = 1
    LED_ROT = 2
    RES_BLE = 3
    BLE_ON = 4
    LED_YELLOW = 5
    BAT_ENABLE = 6
    # led blue and digital out 9 is connected on the pcb
    LED_BLUE = 7
    DOUT9 = 7

    LUEFTER_ON1 = 8
    LUEFTER_ON2 = 9
    LUEFTER_ON3 = 10
    PRESS_ON = 11
    PRESS_DIR = 12
    LIFT_AB = 13
    LIFT_AUF = 14
    DOUT10 = 15

    DOUT1 = 16
    REL_RES = 16

    DOUT2 = 17
    REL_VAKUUM = 17

    DOUT3 = 18
    DOUT4 = 19
    DOUT5 = 20
    DOUT6 = 21
    DOUT7 = 22
    DOUT8 = 23

    # Inputs
    MAIN_SWITCH_IN = 0
    DOOR_END_SWITCH_CLOSED = 1
    DOOR_END_SWITCH_OPEN = 2
    SAFETY_SWITCH = 3
    DIN4 = 4
    DIN5 = 5
    FAN1_ON = 6
    FAN2_ON = 7
    FAN3_ON = 8
    RES_DIG_IN9 = 9
    RES_DIG_IN10 = 10
    RES_DIG_IN11 = 11
    V24_STATUS = 12

    ALL_ON = int(0xffffff)
    ALL_OFF = int(0x000000)

    def __init__(self):
        """!
        Initialisiert die Software SPI
        Definiert die notwendigen Pins
        Setzt die Pins auf die richtigen Zustände (In/Out, High/Low)
        """
        self._lock = threading.Lock()
        self.MISO = mraa.Gpio(5)
        self.MOSI = mraa.Gpio(1)
        self.CLK0 = mraa.Gpio(0)  # clk for hc594
        self.CLK1 = mraa.Gpio(4)  # clk for hc165
        self.CS0 = mraa.Gpio(2)  # load, chip select for hc594
        self.CS1 = mraa.Gpio(3)  # load, chip select for hc165

        # self.callbacks = [None] * 40

        # 40 bit output data, the first two bytes is for the 2 bytes input data
        # self.output = int(0x0000000000)
        # 40 bit input data, the last two bytes represents the 2 byte input data
        # self.input = int(0x0000000000)

        # 24 bit output data
        self._out_value = int(0x000000)
        # 16 bit input data
        self._in_value = int(0x0000)

        self.MOSI.dir(mraa.DIR_OUT)
        self.MISO.dir(mraa.DIR_IN)
        self.CLK0.dir(mraa.DIR_OUT)
        self.CLK1.dir(mraa.DIR_OUT)
        self.CS0.dir(mraa.DIR_OUT)
        self.CS1.dir(mraa.DIR_OUT)

        # hc594
        self.CS0.write(0)
        self.CLK0.write(0)

        # hc165
        self.CS1.write(0)
        self.CLK1.write(0)

        self.update()

    def set_output_bit(self, bit_position):
        """!
        Setzt das Bit an bit_position in _out_value, muss dann per update() übertragen werden.
        @param bit_position: Nummer des Bits, das gesetzt werden soll
        """
        with self._lock:
            if bit_position < 0 or bit_position > 23:
                raise RuntimeError('Bitposition is not correct')
            self._out_value |= (1 << bit_position)

    def clear_output_bit(self, bit_position):
        """!
        Löscht das Bit an bit_position in _out_value, muss dann per update() übertragen werden.
        @param bit_position: Nummer des Bits, das gelöscht werden soll
        """
        with self._lock:
            if bit_position < 0 or bit_position > 23:
                raise RuntimeError('Bitposition is not correct')
            self._out_value &= ~(1 << bit_position)

    def output_is_set(self, bit_position):
        """!
        Gibt True zurück, wenn das Bit an bit_position in _out_value gesetzt ist.
        Achtung: Gibt nur den Status der Variable zurück, nicht den des Ausgangs-Schieberegisters!
        @param bit_position: Nummer des Bits, dessen Zustand geprüft werden soll
        @return True - Bit ist gesetzt, False - Bit ist nicht gesetzt
        """
        with self._lock:
            if bit_position < 0 or bit_position > 23:
                raise RuntimeError('Bitposition is not correct')
            if self._out_value & (1 << bit_position):
                return True
            else:
                return False

    def input_is_set(self, bit_position):
        """!
        Gibt True zurück, wenn das Bit an bit_position in _in_value gesetzt ist.
        Achtung: Gibt den Status des Eingangs-Schieberegisters nur zurück, wenn dieses zuvor mit update() gelesen wurde!
        @param bit_position: Nummer des Bits, dessen Zustand geprüft werden soll
        @return True - Bit ist gesetzt, False - Bit ist nicht gesetzt
        """
        with self._lock:
            if bit_position < 0 or bit_position > 15:
                raise RuntimeError('Bitposition is not correct')
            if self._in_value & (1 << bit_position):
                return True
            else:
                return False

    def set_output(self, output):
        """!
        Setzt _out_value auf den übergebenen Wert, muss dann per update() übertragen werden.
        @param output: Wert, der an das Ausgangs-Schieberegister übertragen werden soll.
        """
        with self._lock:
            self._out_value = output

    def get_input(self):
        """!
        Gibt den Wert des Eingangs-Schieberegisters zurück, nachdem dieses zuvor mit update() gelesen wurde.
        @return Wert des Eingangs-Schieberegisters.
        """
        with self._lock:
            return self._in_value

    def get_output(self):
        """!
        Gibt den aktuellen Wert der Variable _out_value zurück.
        Achtung: Gibt nur den Wert der Variable zurück, nicht den des Ausgangs-Schieberegisters!
        @return Wert des Ausgangsregisters.
        """
        with self._lock:
            return self._out_value

    def update(self):
        """!
        Die Update Funktion schiebt den _out_value in die Ausgangs-Schieberegister.
        Gleichzeitig holt sie den Inhalt der Eingangs-Schieberegister ab und speichert diesen in _in_value.
        Die Funktion generiert hierfür die CS und CLK Signale.
        Die Werte für die Ausgangs-Schieberegister werden über MOSI geschrieben.
        Die Werte für die Eingangs-Schieberegister werden über MISO gelesen.
        """
        # hc594
        self.CLK0.write(1)
        # CS0.write(1)
        # CS0.write(0)

        # hc165
        self.CLK1.write(1)
        self.CS1.write(0)
        self.CS1.write(1)

        # out_val = int(self.output)
        # the most significant bytes are used for the output
        out_val = (int(0x000000) | self._out_value)
        in_val = int(0x000000)

        # print("data_to_send = {0:040b}".format(out_val))
        for index in xrange(24):
            # write bit
            if out_val & 0x800000:
                self.MOSI.write(1)
            else:
                self.MOSI.write(0)
            # next bit
            out_val <<= 1

            # next bit
            in_val <<= 1

            # read input
            if self.MISO.read() is 1:
                in_val |= 0x01

            # print("index:{}, test:{}".format(index, bit))
            # sleep(0.001)
            self.CLK0.write(0)
            self.CLK1.write(0)
            # sleep(0.001)
            self.CLK0.write(1)
            self.CLK1.write(1)

        # hc594
        self.CS0.write(1)
        self.CS0.write(0)

        # hc165
        self.CS1.write(0)
        # CS1.write(1)

        # self.input = in_val
        self._in_value = (0xFFFF & (in_val >> 8))

